/*
** compress.c - Main compression routine for LZA file compression program.
**
** Author: DavidDi
*/

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include "lz_common.h"
#include "lzcommon.h"
#include "lz_buffers.h"
uint8_t *slurp_file(char *path, uint64_t *filesize);

char MakeCompressedName(char *pszFileName)
{
    char chReplaced = '\0';
    size_t NameLength = strlen(pszFileName), DotIndex = NameLength;

    while ((DotIndex > 0) && (pszFileName[--DotIndex] != '.')) {
        if ((pszFileName[DotIndex] == '\\') ||
            (pszFileName[DotIndex] == ':')) {   // end of filename part of path
            DotIndex = 0;                       // name has no extension
            break;
        }
    }

    if (DotIndex > 0) {                         // name has an extension
        if ((NameLength - DotIndex ) <= 3) {    // extension less than 3 chars
            pszFileName[NameLength++] = '_';    // append '_' to extension
            pszFileName[NameLength++] = 0;      // terminate
        } else {                                // extension more than 3 chars
            chReplaced = pszFileName[NameLength - 1]; // return original
            pszFileName[NameLength - 1] = '_';  // replace last char with '_'
        }
    } else {                                    // name has no extension
        pszFileName[NameLength++] = '.';        // append '.'
        pszFileName[NameLength++] = '_';        // append '_'
        pszFileName[NameLength++] = 0;          // terminate
    }

    return(chReplaced);
}

LZINFO *InitGlobalRingbuffer(uint32_t dwRingBufSize)
{
   LZINFO *pLZI;

   if (!(pLZI = (LZINFO*)malloc(sizeof(LZINFO)))) {
      return(NULL);
   }

   // Set up ring buffer.  N.b., extra (cbStrMax - 1) bytes used to
   // facilitate string comparisons near end of ring buffer.
   // (The size allocated for the ring buffer may be at most 4224, since
   //  that's the ring buffer length embedded in the LZFile structs in
   //  lzexpand.h.)
   if (dwRingBufSize == 0) { dwRingBufSize = MAX_RING_BUF_LEN; }

   if ((pLZI->rgbyteRingBuf = (uint8_t*)malloc(dwRingBufSize)) == NULL)
      // Bail out, since without the ring buffer, we can't decode anything.
      return(NULL);
   return(pLZI);
}

/*
** int WriteOutBuf(BYTE byteNext, int doshDest);
**
** Dumps output buffer to output file.  Prompts for new floppy disk if the
** old one if full.  Continues dumping to output file of same name on new
** floppy disk.
**
** Arguments:  byteNext - first byte to be added to empty buffer after buffer
**                        is written
**             doshDest - output DOS file handle
**
** Returns:    int - TRUE if successful.  LZERROR_BADOUTHANDLE or
**                   LZERROR_WRITE if unsuccessful.
**
** Globals:    pbyteOutBuf - reset to point to free byte after byteNext in
**                           rgbyteOutBuf
*/
// !!! Assumes pLZI parm is valid.  No sanity check (should be done above in caller)./
int WriteOutBuf(int doshDest, LZINFO *pLZI)
{
   uint32_t ucbToWrite,       // number of bytes to write from buffer
         ucbWritten;       // number of bytes actually written

   // How much of the buffer should be written to the output file?
   ucbToWrite = (size_t)(pLZI->pbyteOutBuf - pLZI->rgbyteOutBuf);
   // Reset pointer to beginning of buffer.
   pLZI->pbyteOutBuf = pLZI->rgbyteOutBuf;

   // Write to ouput file.
   if (doshDest != -1 &&
       (ucbWritten = write(doshDest, pLZI->pbyteOutBuf, ucbToWrite)) != ucbToWrite) {
      // Insufficient space on destination drive.
      return(LZERROR_WRITE);
   }

   // Add the next byte to the buffer.
   //`*pLZI->pbyteOutBuf++ = byteNext;
   return(TRUE);
}

/*
** int Compress(char *pszSource, char *pszDest,
**              BYTE byteAlgorithm, BYTE byteExtensionChar);
**
** Compress one file to another.
**
** Arguments:  pszSource         - name of file to compress
**             pszDest           - name of compressed output file
**             byteAlgorithm     - compression algorithm to use
**             byteExtensionChar - compressed file name extension character
**
** Returns:    int - TRUE if compression finished successfully.  One of the
**                   LZERROR_ codes if not.
**
** Globals:    none
*/
int LZCompress(char *pszDest, const char *pszSource, int bDoRename)
{
   LZINFO *pLZI = InitGlobalRingbuffer(8192);
   uint64_t filesz = 0;
   int out_fd = -1,                // output file handle
       nRetVal = 1;
   FH FHOut;                  // compressed header info struct
   char szDestFileName[16384];
   uint8_t byteAlgorithm, byteExtensionChar;

   // Try to set up input buffer.  N.b., extra byte because rgbyteInBuf[0]
   // will be used to hold last byte from previous input buffer.
   pLZI->rgbyteInBuf = slurp_file(pszSource, &filesz);
   pLZI->ucbInBufLen = filesz;
   pLZI->cblInSize = filesz;
   pLZI->ucbOutBufLen = MAX_OUT_BUF_SIZE;

   // Create destination file name.
   strcpy(szDestFileName, pszDest);

   if (bDoRename)
      // Rename output file.
      byteExtensionChar = MakeCompressedName(szDestFileName);
   else
      byteExtensionChar = '\0';

   //out_fd = open(pszDest, O_WRONLY|O_CREAT|O_TRUNC, 0644);
   out_fd = 1;

   // Set cblInSize to length of input file.
   byteAlgorithm = ALG_FIRST;
   //pLZI->cblInSize = st.st_size;
   // Fill in compressed file header.
   MakeHeader(&FHOut, byteAlgorithm, byteExtensionChar, pLZI);
   // Write compressed file header to output file.
   write(out_fd, &FHOut, sizeof(FH)-sizeof(uint32_t));
   // Compress input file into output file.
   nRetVal = LZEncode(out_fd, pszSource, pLZI);
   // Copy date and time stamp from source file to destination file.
   //CopyDateTimeStamp(szDestFileName, pszSource);

COMPRESS_EXIT:
   // Close files.
   //close (out_fd);
   return(nRetVal);
}

int main(int argc, char **argv) {
    LZCompress(argv[1], argv[1], 0);
    //write(1, outbuf, lzs.reallength);
    return 0;
}

