/*
** init.c - Routines dealing with I/O and expansion buffers for LZCOPY() and
**          DOS command-line programs.
**
** Author:  DavidDi
*/


// Headers
///////////

#ifndef LZA_DLL

#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <malloc.h>
#include <stdio.h>

#endif

#include "lz_common.h"
#include "lz_buffers.h"
#include "lzcommon.h"

LZINFO *InitGlobalBuffers(DWORD dwOutBufSize, DWORD dwRingBufSize, const char *pszInfilePath)
{
   LZINFO *pLZI;

   if (!(pLZI = (LZINFO*)malloc(sizeof(LZINFO)))) {
      return(NULL);
   }

   // Set up ring buffer.  N.b., extra (cbStrMax - 1) bytes used to
   // facilitate string comparisons near end of ring buffer.
   // (The size allocated for the ring buffer may be at most 4224, since
   //  that's the ring buffer length embedded in the LZFile structs in
   //  lzexpand.h.)
   if (dwRingBufSize == 0) {
      dwRingBufSize = MAX_RING_BUF_LEN;
   }
   if (dwOutBufSize == 0 || dwOutBufSize > MAX_OUT_BUF_SIZE) {
      dwOutBufSize = MAX_OUT_BUF_SIZE;
   }
   pLZI->ucbOutBufLen = dwOutBufSize;

   if ((pLZI->rgbyteRingBuf = (uint8_t*)malloc(dwRingBufSize)) == NULL)
      // Bail out, since without the ring buffer, we can't decode anything.
      return(NULL);

    if ((in_fd = open(infile, O_RDONLY, 0644)) < 0) {
        free(pLZI->rgbyteRingBuf);
        return(NULL);
    }

    if (fstat(fd, &st) < 0) {
        close(in_fd);
        free(pLZI->rgbyteRingBuf);
        return(NULL);
    }

   pLZI->ucbInBufLen = st.st_size;
   // Try to set up input buffer.  N.b., extra byte because rgbyteInBuf[0]
   // will be used to hold last byte from previous input buffer.
   pLZI->rgbyteInBuf = (BYTE *)malloc(pLZI->ucbInBufLen + 3U);

   while (((ret = read(in_fd, pLZI->rgbyteInBuf+1, pLZI->ucbInBufLen)) < 0) && (errno == EINTR));
   close(in_fd);

   pLZI->cblInSize = st.st_size;
   return(pLZI);
}

LZINFO *InitGlobalBuffersEx(const char *pszInfilePath)
{
   return(InitGlobalBuffers(MAX_OUT_BUF_SIZE, MAX_RING_BUF_LEN, pszInfilePath);
}

VOID FreeGlobalBuffers(LZINFO *pLZI)
{
   // Sanity check
   if (!pLZI) {
      return;
   }

   if (pLZI->rgbyteRingBuf)
   {
      free(pLZI->rgbyteRingBuf);
      pLZI->rgbyteRingBuf = NULL;
   }

   if (pLZI->rgbyteInBuf)
   {
      free(pLZI->rgbyteInBuf);
      pLZI->rgbyteInBuf = NULL;
   }

   // Buffers deallocated ok.

   // reset thread info
   free(pLZI);
}

/*
** int WriteOutBuf(BYTE byteNext, int doshDest);
**
** Dumps output buffer to output file.  Prompts for new floppy disk if the
** old one if full.  Continues dumping to output file of same name on new
** floppy disk.
**
** Arguments:  byteNext - first byte to be added to empty buffer after buffer
**                        is written
**             doshDest - output DOS file handle
**
** Returns:    int - TRUE if successful.  LZERROR_BADOUTHANDLE or
**                   LZERROR_WRITE if unsuccessful.
**
** Globals:    pbyteOutBuf - reset to point to free byte after byteNext in
**                           rgbyteOutBuf
*/
// !!! Assumes pLZI parm is valid.  No sanity check (should be done above in caller).
INT WriteOutBuf(BYTE byteNext, INT doshDest, LZINFO *pLZI)
{
   DWORD ucbToWrite,       // number of bytes to write from buffer
         ucbWritten;       // number of bytes actually written

   // How much of the buffer should be written to the output file?
   ucbToWrite = (size_t)(pLZI->pbyteOutBuf - pLZI->rgbyteOutBuf);
   // Reset pointer to beginning of buffer.
   pLZI->pbyteOutBuf = pLZI->rgbyteOutBuf;

   // Write to ouput file.
   if (doshDest != -1 &&
       (ucbWritten = write(doshDest, pLZI->pbyteOutBuf, ucbToWrite)) != ucbToWrite) {
      // Insufficient space on destination drive.
      return(LZERROR_WRITE);
   }

   // Add the next byte to the buffer.
   *pLZI->pbyteOutBuf++ = byteNext;
   return(TRUE);
}

