/*
** common.h - housekeeping for Lempel-Ziv compression / expansion DOS
**            command-line programs, DOS static library module, and Windows
**            DLL
**
** Author:  DavidDi
*/


// Headers
///////////

//#include <lzdos.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
//#include "translat.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

// Set up type for function argument pointers.
#define ARG_PTR


/////////////
// Constants
/////////////

// length of entire compressed file header (used as offset to start of compressed data)
// (14 == cbCompSigLength + algorithm + extension character + uncompressed length)
#define HEADER_LEN      14

// compressed file signature: "SZDD??'3"
#define COMP_SIG        "SZDD\x88\xf0\x27\x33"
#define COMP_SIG_LEN    8              // length of signature (bytes) (no '\0' terminator)

#define ALG_FIRST       ((uint8_t) 'A')   // first version algorithm label for Lempel-Ziv
#define ALG_LZ          ((uint8_t) 'B')   // new Lempel-Ziv algorithm label
#define ALG_LZA         ((uint8_t) 'C')   // Lempel-Ziv with arithmetic encoding algorithm label

#define chEXTENSION_CHAR      '_'
#define pszEXTENSION_STR      "_"
#define pszNULL_EXTENSION     "._"

#define LZERROR_OK 0
#define LZERROR_INVALID_ARG 1
#define LZERROR_BADINBUFFER 2
#define LZERROR_WRITE 3
#define LZERROR_READ 3
#define BLANK_ERROR           0        // error condition requiring no error message display


/////////
// Types
/////////


// Flag indicating whether or not rgbyteInBuf[0], which holds the last byte
// from the previous input buffer, should be read as the next input byte.
// (Only used so that at least one unReadUChar() can be done at all input
// buffer positions.)

typedef struct tagLZI {
   uint8_t *rgbyteRingBuf;  // ring buffer for expansion
   uint8_t *rgbyteInBuf;    // input buffer for reads
   uint8_t *pbyteInBufEnd;  // pointer past end of rgbyteInBuf[]
   uint8_t *pbyteInBuf;     // pointer to next byte to read from
   uint8_t  rgbyteOutBuf[4096];   // output buffer for writes
   uint8_t *pbyteOutBufEnd; // pointer past end of rgbyteOutBuf[]
   uint8_t *pbyteOutBuf;    // pointer to last byte to write from
   // Flag indicating whether or not rgbyteInBuf[0], which holds the last byte
   // from the previous input buffer, should be read as the next input byte.
   // (Only used so that at least one unReadUChar() can be done at all input
   // buffer positions.)
   uint8_t bLastUsed;
   // Actually, rgbyteInBuf[] has length (ucbInBufLen + 1) since rgbyteInBuf[0]
   // is used when bLastUsed is TRUE.
   int cbMaxMatchLen;     // longest match length for current algorithm
   off_t cblInSize,       // size in bytes of input file
         cblOutSize;      // size in bytes of output file
   uint32_t ucbInBufLen,  // length of input buffer
            ucbOutBufLen; // length of output buffer
   uint32_t uFlags;       // LZ decoding description byte
   int iCurRingBufPos;    // ring buffer offset
   int *leftChild;        // parents and left and right
   int *rightChild;       // children that make up the
   int *parent;           // binary search trees

   int iCurMatch,         // index of longest match (set by LZInsertNode())
       cbCurMatch;        // length of longest match (set by LZInsertNode())

} LZINFO;

// Declare compressed file header information structure.  N.b., the
// compressed file header does not contain the file size of the compressed
// file since this is readily obtainable through filelength() or lseek().
// The file info structure, however, does contain the compressed file size,
// which is used when expanding the file.
static const uint8_t LZMagic[COMP_SIG_LEN]={'S','Z','D','D',0x88,0xf0,0x27,0x33};

typedef struct tagFH
{
   uint8_t rgbyteMagic[COMP_SIG_LEN];  // array of compressed file signature (magic bytes)

   uint8_t byteAlgorithm;           // algorithm label
   char    byteExtensionChar;       // last extension character (always 0 for ALG_FIRST)

   // The file sizes are unsigned longs instead of signed longs for backward
   // compatibilty with version 1.00.
   uint32_t cbulUncompSize;    // uncompressed file size
   uint32_t cbulCompSize;      // compressed file size (not stored in header)
} __attribute__((packed)) FH;


//////////
// Macros
//////////

#define FOREVER   for(;;)

#ifndef MAX
#define MAX(a, b)             (((a) > (b)) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a, b)             (((a) < (b)) ? (a) : (b))
#endif

#if 0
#define RecognizeCompAlg(chAlg)  ((chAlg) == ALG_FIRST || (chAlg) == ALG_LZ || (chAlg) == ALG_LZA)
#else
#define RecognizeCompAlg(chAlg)  ((chAlg) == ALG_FIRST)
#endif


////////////////////
// Inline functions 
///////////////////

/*
** Determine if a file is compresssed using LZA.
** Returns 1 if file signature matches expected compressed file signature, 0 otherwise.
*/
static inline int IsCompressed(FH *pFHIn)
{
   return(!memcmp(pFHIn->rgbyteMagic, LZMagic, 8));
}

/*
** Global cblInSize is used to fill in expanded file length field.
** Compressed file length field is set to 0 since it isn't written.
** !!! Assumes pLZI parm is valid.  No sanity check (should be done above in caller).
*/
static inline void MakeHeader(FH *pFHBlank, uint8_t byteAlgorithm,
                              uint8_t byteExtensionChar, LZINFO *pLZI)
{
   // Fill in compressed file signature.
   memcpy(pFHBlank->rgbyteMagic, LZMagic, COMP_SIG_LEN);
   // Fill in algorithm and extesion character.
   pFHBlank->byteAlgorithm = byteAlgorithm;
   pFHBlank->byteExtensionChar = byteExtensionChar;
   // Fill in file sizes.  (cbulCompSize not written to compressed file
   // header, so just set it to 0UL.)
   pFHBlank->cbulUncompSize = (uint32_t)pLZI->cblInSize;
   pFHBlank->cbulCompSize = 0UL;
}


//////////////
// Prototypes
//////////////

#if 0
// compress.c
extern int Compress(CHAR *pszSource,
                    CHAR *pszDest, uint8_t byteAlgorithm,
                    BOOL bDoRename, LZINFO *pLZI);

// expand.c
extern int Expand(CHAR *pszSource, CHAR *pszDest, BOOL bDoRename, LZINFO *pLZI);
extern int ExpandOrCopyFile(int doshSource, int doshDest, PLZINFO pLZI);

// header.c
extern int WriteHdr(FH *pFH, int doshDest, LZINFO *pLZI);
extern int GetHdr(FH *pFH, int doshSource, uint32_t *pcblInSize);

// dosdir.asm
extern int GetCurDir(LPSTR lpszDirBuf);
extern int SetDir(LPSTR lpszDirName);
extern int IsDir(LPSTR lpszDir);

// utils.c
extern char *ExtractFileName(char *pszPathName);
extern char *ExtractExtension(char *pszFileName);
extern void MakePathName(char *pszPath, char *pszFileName);
extern CHAR MakeCompressedName(char *pszFileName);
extern LPWSTR ExtractFileNameW(LPWSTR pszPathName);
extern LPWSTR ExtractExtensionW(LPWSTR pszFileName);
extern WCHAR MakeCompressedNameW(LPWSTR pszFileName);
extern void MakeExpandedName(CHAR *pszFileName, uint8_t byteExtensionChar);
extern int CopyDateTimeStamp(int_PTR doshFrom, int_PTR doshTo);

extern int   LZIsCharLowerA(char cChar);
extern int   LZIsCharUpperA(char cChar);
extern char *LZCharNextA(const char *lpCurrentChar);
extern char *LZCharPrevA(const char *lpStart, const char *lpCurrentChar);

#undef IsCharLower
#undef IsCharUpper
#undef CharNext
#undef CharPrev
#define IsCharLower LZIsCharLowerA
#define IsCharUpper LZIsCharUpperA
#define CharNext    LZCharNextA
#define CharPrev    LZCharPrevA
#endif

